@@ -1,24 +1,30 @@
 # Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
-# For details: https://bitbucket.org/ned/coveragepy/src/default/NOTICE.txt
+# For details: https://github.com/nedbat/coveragepy/blob/master/NOTICE.txt
 
 """Oddball cases for testing coverage.py"""
 
+import os.path
+import re
 import sys
 
+from flaky import flaky
 import pytest
 
 import coverage
 from coverage import env
+from coverage.data import sorted_lines
 from coverage.files import abs_file
+from coverage.misc import import_local_file
 
 from tests.coveragetest import CoverageTest
+from tests.helpers import swallow_warnings
 from tests import osinfo
 
 
 class ThreadingTest(CoverageTest):
     """Tests of the threading support."""
 
-    def test_threading(self):
+    def test_threading(self) -> None:
         self.check_coverage("""\
             import threading
 
@@ -38,7 +44,7 @@ class ThreadingTest(CoverageTest):
             """,
             [1, 3, 4, 6, 7, 9, 10, 12, 13, 14, 15], "10")
 
-    def test_thread_run(self):
+    def test_thread_run(self) -> None:
         self.check_coverage("""\
             import threading
 
@@ -61,7 +67,7 @@ class ThreadingTest(CoverageTest):
 class RecursionTest(CoverageTest):
     """Check what happens when recursive code gets near limits."""
 
-    def test_short_recursion(self):
+    def test_short_recursion(self) -> None:
         # We can definitely get close to 500 stack frames.
         self.check_coverage("""\
             def recur(n):
@@ -75,23 +81,24 @@ class RecursionTest(CoverageTest):
             """,
             [1, 2, 3, 5, 7, 8], "")
 
-    def test_long_recursion(self):
+    def test_long_recursion(self) -> None:
         # We can't finish a very deep recursion, but we don't crash.
-        with self.assertRaises(RuntimeError):
-            self.check_coverage("""\
-                def recur(n):
-                    if n == 0:
-                        return 0
-                    else:
-                        return recur(n-1)+1
+        with pytest.raises(RuntimeError):
+            with swallow_warnings("Trace function changed, data is likely wrong: None"):
+                self.check_coverage("""\
+                    def recur(n):
+                        if n == 0:
+                            return 0
+                        else:
+                            return recur(n-1)+1
 
-                recur(100000)  # This is definitely too many frames.
-                """,
-                [1, 2, 3, 5, 7], ""
+                    recur(100000)  # This is definitely too many frames.
+                    """,
+                    [1, 2, 3, 5, 7], ""
                 )
 
-    def test_long_recursion_recovery(self):
-        # Test the core of bug 93: http://bitbucket.org/ned/coveragepy/issue/93
+    def test_long_recursion_recovery(self) -> None:
+        # Test the core of bug 93: https://github.com/nedbat/coveragepy/issues/93
         # When recovering from a stack overflow, the Python trace function is
         # disabled, but the C trace function is not.  So if we're using a
         # Python trace function, we won't trace anything after the stack
@@ -114,39 +121,44 @@ class RecursionTest(CoverageTest):
             """)
 
         cov = coverage.Coverage()
-        self.start_import_stop(cov, "recur")
+        with swallow_warnings("Trace function changed, data is likely wrong: None"):
+            self.start_import_stop(cov, "recur")
 
-        pytrace = (cov.collector.tracer_name() == "PyTracer")
+        assert cov._collector is not None
+        pytrace = (cov._collector.tracer_name() == "PyTracer")
         expected_missing = [3]
-        if pytrace:
+        if pytrace:                                 # pragma: no metacov
             expected_missing += [9, 10, 11]
 
         _, statements, missing, _ = cov.analysis("recur.py")
-        self.assertEqual(statements, [1, 2, 3, 5, 7, 8, 9, 10, 11])
-        self.assertEqual(missing, expected_missing)
+        assert statements == [1, 2, 3, 5, 7, 8, 9, 10, 11]
+        assert expected_missing == missing
 
         # Get a warning about the stackoverflow effect on the tracing function.
-        if pytrace:
-            self.assertEqual(cov._warnings,
-                ["Trace function changed, measurement is likely wrong: None"]
-                )
+        if pytrace:                                 # pragma: no metacov
+            assert len(cov._warnings) == 1
+            assert re.fullmatch(
+                r"Trace function changed, data is likely wrong: None != " +
+                r"<bound method PyTracer._trace of " +
+                "<PyTracer at 0x[0-9a-fA-F]+: 5 data points in 1 files>>",
+                cov._warnings[0],
+            )
         else:
-            self.assertEqual(cov._warnings, [])
+            assert not cov._warnings
 
 
 class MemoryLeakTest(CoverageTest):
     """Attempt the impossible: test that memory doesn't leak.
 
     Note: this test is truly unusual, and has had a colorful history.  See
-    for example: https://bitbucket.org/ned/coveragepy/issue/186
+    for example: https://github.com/nedbat/coveragepy/issues/186
 
     It may still fail occasionally, especially on PyPy.
 
     """
-    def test_for_leaks(self):
-        if env.JYTHON:
-            self.skipTest("Don't bother on Jython")
-
+    @flaky      # type: ignore[misc]
+    @pytest.mark.skipif(not env.C_TRACER, reason="Only the C tracer has refcounting issues")
+    def test_for_leaks(self) -> None:
         # Our original bad memory leak only happened on line numbers > 255, so
         # make a code object with more lines than that.  Ugly string mumbo
         # jumbo to get 300 blank lines at the beginning..
@@ -186,19 +198,19 @@ class MemoryLeakTest(CoverageTest):
                 fails += 1                                  # pragma: only failure
 
         if fails > 8:
-            self.fail("RAM grew by %d" % (ram_growth))      # pragma: only failure
+            pytest.fail("RAM grew by %d" % (ram_growth))      # pragma: only failure
 
 
 class MemoryFumblingTest(CoverageTest):
     """Test that we properly manage the None refcount."""
 
-    def test_dropping_none(self):                           # pragma: not covered
-        if not env.C_TRACER:
-            self.skipTest("Only the C tracer has refcounting issues")
+    @pytest.mark.skipif(not env.C_TRACER, reason="Only the C tracer has refcounting issues")
+    def test_dropping_none(self) -> None:                     # pragma: not covered
         # TODO: Mark this so it will only be run sometimes.
-        self.skipTest("This is too expensive for now (30s)")
+        pytest.skip("This is too expensive for now (30s)")
         # Start and stop coverage thousands of times to flush out bad
         # reference counting, maybe.
+        _ = "this is just here to put a type comment on"    # type: ignore[unreachable]
         self.make_file("the_code.py", """\
             import random
             def f():
@@ -220,15 +232,15 @@ class MemoryFumblingTest(CoverageTest):
             print("Final None refcount: %d" % (sys.getrefcount(None)))
             """)
         status, out = self.run_command_status("python main.py")
-        self.assertEqual(status, 0)
-        self.assertIn("Final None refcount", out)
-        self.assertNotIn("Fatal", out)
+        assert status == 0
+        assert "Final None refcount" in out
+        assert "Fatal" not in out
 
 
 class PyexpatTest(CoverageTest):
     """Pyexpat screws up tracing. Make sure we've counter-defended properly."""
 
-    def test_pyexpat(self):
+    def test_pyexpat(self) -> None:
         # pyexpat calls the trace function explicitly (inexplicably), and does
         # it wrong for exceptions.  Parsing a DOCTYPE for some reason throws
         # an exception internally, and triggers its wrong behavior.  This test
@@ -261,20 +273,18 @@ class PyexpatTest(CoverageTest):
         self.start_import_stop(cov, "outer")
 
         _, statements, missing, _ = cov.analysis("trydom.py")
-        self.assertEqual(statements, [1, 3, 8, 9, 10, 11, 13])
-        self.assertEqual(missing, [])
+        assert statements == [1, 3, 8, 9, 10, 11, 13]
+        assert missing == []
 
         _, statements, missing, _ = cov.analysis("outer.py")
-        self.assertEqual(statements, [101, 102])
-        self.assertEqual(missing, [])
+        assert statements == [101, 102]
+        assert missing == []
 
         # Make sure pyexpat isn't recorded as a source file.
-        # https://bitbucket.org/ned/coveragepy/issues/419/nosource-no-source-for-code-path-to-c
+        # https://github.com/nedbat/coveragepy/issues/419
         files = cov.get_data().measured_files()
-        self.assertFalse(
-            any(f.endswith("pyexpat.c") for f in files),
-            "Pyexpat.c is in the measured files!: %r:" % (files,)
-        )
+        msg = f"Pyexpat.c is in the measured files!: {files!r}:"
+        assert not any(f.endswith("pyexpat.c") for f in files), msg
 
 
 class ExceptionTest(CoverageTest):
@@ -282,7 +292,7 @@ class ExceptionTest(CoverageTest):
     in the trace function.
     """
 
-    def test_exception(self):
+    def test_exception(self) -> None:
         # Python 2.3's trace function doesn't get called with "return" if the
         # scope is exiting due to an exception.  This confounds our trace
         # function which relies on scope announcements to track which files to
@@ -328,7 +338,7 @@ class ExceptionTest(CoverageTest):
         # Import all the modules before starting coverage, so the def lines
         # won't be in all the results.
         for mod in "oops fly catch doit".split():
-            self.import_local_file(mod)
+            import_local_file(mod)
 
         # Each run nests the functions differently to get different
         # combinations of catching exceptions and letting them fly.
@@ -337,31 +347,31 @@ class ExceptionTest(CoverageTest):
                 'doit.py': [302, 303, 304, 305],
                 'fly.py': [102, 103],
                 'oops.py': [2, 3],
-                }),
+            }),
             ("doit catch oops", {
                 'doit.py': [302, 303],
                 'catch.py': [202, 203, 204, 206, 207],
                 'oops.py': [2, 3],
-                }),
+            }),
             ("doit fly catch oops", {
                 'doit.py': [302, 303],
                 'fly.py': [102, 103, 104],
                 'catch.py': [202, 203, 204, 206, 207],
                 'oops.py': [2, 3],
-                }),
+            }),
             ("doit catch fly oops", {
                 'doit.py': [302, 303],
                 'catch.py': [202, 203, 204, 206, 207],
                 'fly.py': [102, 103],
                 'oops.py': [2, 3],
-                }),
-            ]
+            }),
+        ]
 
         for callnames, lines_expected in runs:
 
             # Make the list of functions we'll call for this test.
-            callnames = callnames.split()
-            calls = [getattr(sys.modules[cn], cn) for cn in callnames]
+            callnames_list = callnames.split()
+            calls = [getattr(sys.modules[cn], cn) for cn in callnames_list]
 
             cov = coverage.Coverage()
             cov.start()
@@ -374,32 +384,24 @@ class ExceptionTest(CoverageTest):
             # The file names are absolute, so keep just the base.
             clean_lines = {}
             data = cov.get_data()
-            for callname in callnames:
+            for callname in callnames_list:
                 filename = callname + ".py"
-                lines = data.lines(abs_file(filename))
-                clean_lines[filename] = sorted(lines)
+                clean_lines[filename] = sorted_lines(data, abs_file(filename))
 
-            self.assertEqual(clean_lines, lines_expected)
+            assert clean_lines == lines_expected
 
 
 class DoctestTest(CoverageTest):
     """Tests invoked with doctest should measure properly."""
 
-    def setUp(self):
-        super(DoctestTest, self).setUp()
-
-        # Oh, the irony!  This test case exists because Python 2.4's
-        # doctest module doesn't play well with coverage.  But nose fixes
-        # the problem by monkeypatching doctest.  I want to undo the
-        # monkeypatch to be sure I'm getting the doctest module that users
-        # of coverage will get.  Deleting the imported module here is
-        # enough: when the test imports doctest again, it will get a fresh
-        # copy without the monkeypatch.
-        if 'doctest' in sys.modules:
-            del sys.modules['doctest']
-
-    def test_doctest(self):
-        self.check_coverage('''\
+    def test_doctest(self) -> None:
+        # Doctests used to be traced, with their line numbers credited to the
+        # file they were in.  Below, one of the doctests has four lines (1-4),
+        # which would incorrectly claim that lines 1-4 of the file were
+        # executed.  In this file, line 2 is not executed.
+        self.make_file("the_doctest.py", '''\
+            if "x" in "abc":
+                print("hello")
             def return_arg_or_void(arg):
                 """If <arg> is None, return "Void"; otherwise return <arg>
 
@@ -409,6 +411,11 @@ class DoctestTest(CoverageTest):
                 'arg'
                 >>> return_arg_or_void("None")
                 'None'
+                >>> if "x" in "xyz":                # line 1
+                ...   if "a" in "aswed":            # line 2
+                ...      if "a" in "abc":           # line 3
+                ...         return_arg_or_void(12)  # line 4
+                12
                 """
                 if arg is None:
                     return "Void"
@@ -417,49 +424,59 @@ class DoctestTest(CoverageTest):
 
             import doctest, sys
             doctest.testmod(sys.modules[__name__])  # we're not __main__ :(
-            ''',
-            [1, 11, 12, 14, 16, 17], "")
+            ''')
+        cov = coverage.Coverage()
+        self.start_import_stop(cov, "the_doctest")
+        data = cov.get_data()
+        assert len(data.measured_files()) == 1
+        lines = sorted_lines(data, data.measured_files().pop())
+        assert lines == [1, 3, 18, 19, 21, 23, 24]
 
 
 class GettraceTest(CoverageTest):
     """Tests that we work properly with `sys.gettrace()`."""
-    def test_round_trip(self):
-        self.check_coverage('''\
+    def test_round_trip_in_untraced_function(self) -> None:
+        # https://github.com/nedbat/coveragepy/issues/575
+        self.make_file("main.py", """import sample""")
+        self.make_file("sample.py", """\
+            from swap import swap_it
+            def doit():
+                print(3)
+                swap_it()
+                print(5)
+            def doit_soon():
+                print(7)
+                doit()
+                print(9)
+            print(10)
+            doit_soon()
+            print(12)
+            """)
+        self.make_file("swap.py", """\
             import sys
-            def foo(n):
-                return 3*n
-            def bar(n):
-                return 5*n
-            a = foo(6)
-            sys.settrace(sys.gettrace())
-            a = bar(8)
-            ''',
-            [1, 2, 3, 4, 5, 6, 7, 8], "")
-
-    def test_multi_layers(self):
-        self.check_coverage('''\
-            import sys
-            def level1():
-                a = 3
-                level2()
-                b = 5
-            def level2():
-                c = 7
+            def swap_it():
                 sys.settrace(sys.gettrace())
-                d = 9
-            e = 10
-            level1()
-            f = 12
-            ''',
-            [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12], "")
+            """)
 
-    def test_setting_new_trace_function(self):
-        # https://bitbucket.org/ned/coveragepy/issues/436/disabled-coverage-ctracer-may-rise-from
+        # Use --source=sample to prevent measurement of swap.py.
+        cov = coverage.Coverage(source=["sample"])
+        self.start_import_stop(cov, "main")
+
+        assert self.stdout() == "10\n7\n3\n5\n9\n12\n"
+
+        _, statements, missing, _ = cov.analysis("sample.py")
+        assert statements == [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]
+        assert missing == []
+
+    def test_setting_new_trace_function(self) -> None:
+        # https://github.com/nedbat/coveragepy/issues/436
         self.check_coverage('''\
+            import os.path
             import sys
 
             def tracer(frame, event, arg):
-                print("%s: %s @ %d" % (event, frame.f_code.co_filename, frame.f_lineno))
+                filename = os.path.basename(frame.f_code.co_filename)
+                print(f"{event}: {filename} @ {frame.f_lineno}")
                 return tracer
 
             def begin():
@@ -469,36 +486,35 @@ class GettraceTest(CoverageTest):
                 t = sys.gettrace()
                 assert t is tracer, t
 
-            def test_unsets_trace():
+            def test_unsets_trace() -> None:
                 begin()
                 collect()
 
             old = sys.gettrace()
             test_unsets_trace()
             sys.settrace(old)
+            a = 21
+            b = 22
             ''',
-            lines=[1, 3, 4, 5, 7, 8, 10, 11, 12, 14, 15, 16, 18, 19, 20],
-            missing="4-5, 11-12",
+            lines=[1, 2, 4, 5, 6, 7, 9, 10, 12, 13, 14, 16, 17, 18, 20, 21, 22, 23, 24],
+            missing="5-7, 13-14",
         )
 
+        assert self.last_module_name is not None
         out = self.stdout().replace(self.last_module_name, "coverage_test")
-
-        self.assertEqual(
-            out,
-            (
-                "call: coverage_test.py @ 10\n"
-                "line: coverage_test.py @ 11\n"
-                "line: coverage_test.py @ 12\n"
-                "return: coverage_test.py @ 12\n"
-            ),
+        expected = (
+            "call: coverage_test.py @ 12\n" +
+            "line: coverage_test.py @ 13\n" +
+            "line: coverage_test.py @ 14\n" +
+            "return: coverage_test.py @ 14\n"
         )
+        assert expected == out
 
     @pytest.mark.expensive
-    def test_atexit_gettrace(self):             # pragma: not covered
+    @pytest.mark.skipif(env.METACOV, reason="Can't set trace functions during meta-coverage")
+    def test_atexit_gettrace(self) -> None:
         # This is not a test of coverage at all, but of our understanding
         # of this edge-case behavior in various Pythons.
-        if env.METACOV:
-            self.skipTest("Can't set trace functions during meta-coverage")
 
         self.make_file("atexit_gettrace.py", """\
             import atexit, sys
@@ -517,26 +533,26 @@ class GettraceTest(CoverageTest):
             # This will show what the trace function is at the end of the program.
             """)
         status, out = self.run_command_status("python atexit_gettrace.py")
-        self.assertEqual(status, 0)
+        assert status == 0
         if env.PYPY and env.PYPYVERSION >= (5, 4):
             # Newer PyPy clears the trace function before atexit runs.
-            self.assertEqual(out, "None\n")
+            assert out == "None\n"
         else:
             # Other Pythons leave the trace function in place.
-            self.assertEqual(out, "trace_function\n")
+            assert out == "trace_function\n"
 
 
 class ExecTest(CoverageTest):
     """Tests of exec."""
-    def test_correct_filename(self):
-        # https://bitbucket.org/ned/coveragepy/issues/380/code-executed-by-exec-excluded-from
+    def test_correct_filename(self) -> None:
+        # https://github.com/nedbat/coveragepy/issues/380
         # Bug was that exec'd files would have their lines attributed to the
         # calling file.  Make two files, both with ~30 lines, but no lines in
         # common.  Line 30 in to_exec.py was recorded as line 30 in main.py,
         # but now it's fixed. :)
         self.make_file("to_exec.py", """\
             \n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n
-            print("var is {0}".format(var))         # line 31
+            print("var is {}".format(var))         # line 31
             """)
         self.make_file("main.py", """\
             namespace = {'var': 17}
@@ -551,27 +567,36 @@ class ExecTest(CoverageTest):
         self.start_import_stop(cov, "main")
 
         _, statements, missing, _ = cov.analysis("main.py")
-        self.assertEqual(statements, [1, 2, 3, 4, 35])
-        self.assertEqual(missing, [])
+        assert statements == [1, 2, 3, 4, 35]
+        assert missing == []
         _, statements, missing, _ = cov.analysis("to_exec.py")
-        self.assertEqual(statements, [31])
-        self.assertEqual(missing, [])
+        assert statements == [31]
+        assert missing == []
+
+    def test_unencodable_filename(self) -> None:
+        # https://github.com/nedbat/coveragepy/issues/891
+        self.make_file("bug891.py", r"""exec(compile("pass", "\udcff.py", "exec"))""")
+        cov = coverage.Coverage()
+        self.start_import_stop(cov, "bug891")
+        # Saving would fail trying to encode \udcff.py
+        cov.save()
+        files = [os.path.basename(f) for f in cov.get_data().measured_files()]
+        assert "bug891.py" in files
 
 
 class MockingProtectionTest(CoverageTest):
     """Tests about protecting ourselves from aggressive mocking.
 
-    https://bitbucket.org/ned/coveragepy/issues/416/coverage-40-is-causing-existing-unit-tests
+    https://github.com/nedbat/coveragepy/issues/416
 
     """
-    def test_os_path_exists(self):
+    def test_os_path_exists(self) -> None:
         # To see if this test still detects the problem, change isolate_module
         # in misc.py to simply return its argument.  It should fail with a
         # StopIteration error.
         self.make_file("bug416.py", """\
             import os.path
-
-            import mock
+            from unittest import mock
 
             @mock.patch('os.path.exists')
             def test_path_exists(mock_exists):
@@ -591,4 +616,4 @@ class MockingProtectionTest(CoverageTest):
         import py_compile
         py_compile.compile("bug416a.py")
         out = self.run_command("coverage run bug416.py")
-        self.assertEqual(out, "in test\nbug416a.py\n23\n17\n")
+        assert out == "in test\nbug416a.py\n23\n17\n"
